"""Client API for authentication in Nexus."""

import getpass
import time
import webbrowser
from http import HTTPStatus

import httpx
from colorama import Fore
from pydantic import EmailStr
from rich.console import Console
from rich.panel import Panel

import qnexus.exceptions as qnx_exc
from qnexus.client import (
    VERSION,
    VERSION_HEADER,
    _check_version_headers,
    get_nexus_client,
)
from qnexus.client.utils import consolidate_error, remove_token, write_token
from qnexus.config import CONFIG

console = Console()


def _get_auth_client() -> httpx.Client:
    """Getter function for the Nexus auth client."""
    return httpx.Client(
        base_url=f"{CONFIG.url}/auth",
        timeout=None,
        verify=CONFIG.httpx_verify,
    )


def login() -> None:
    """
    Log in to Quantinuum Nexus using the web browser.

    (if web browser can't be launched, displays the link)
    """

    res = _get_auth_client().post(
        "/device/device_authorization",
        headers={"Content-Type": "application/x-www-form-urlencoded"},
        data={"client_id": "scales", "scope": "myqos"},
    )

    user_code = res.json()["user_code"]
    device_code = res.json()["device_code"]
    verification_uri_complete = res.json()["verification_uri_complete"]
    expires_in = res.json()["expires_in"]
    poll_interval = res.json()["interval"]

    webbrowser.open(verification_uri_complete, new=2)

    token_request_body = {
        "grant_type": "urn:ietf:params:oauth:grant-type:device_code",
        "device_code": device_code,
        "client_id": "scales",
    }

    print("🌐 Browser log in initiated.")

    console.print(
        Panel(
            f"""
        Confirm that the browser shows the following code and click 'allow device':

                                     {user_code}
        """,
            width=90,
        )
    )

    print(
        "Browser didn't open automatically? Use this link: "
        f"{Fore.BLUE + verification_uri_complete}"
    )

    polling_for_seconds = 0
    while polling_for_seconds < expires_in:
        time.sleep(poll_interval)
        polling_for_seconds += poll_interval
        resp = _get_auth_client().post(
            "/device/token",
            headers={
                "Content-Type": "application/x-www-form-urlencoded",
                VERSION_HEADER: VERSION,
            },
            data=token_request_body,
        )
        if (
            resp.status_code == HTTPStatus.BAD_REQUEST
            and resp.json().get("error") == "AUTHORIZATION_PENDING"
        ):
            continue
        if resp.status_code == HTTPStatus.TOO_MANY_REQUESTS:
            continue
        if resp.status_code == HTTPStatus.OK:
            resp_json = resp.json()
            write_token("refresh_token", resp_json["refresh_token"])
            write_token(
                "access_token",
                resp_json["access_token"],
            )
            get_nexus_client(reload=True)
            # spinner.stop()
            print(
                f"✅ Successfully logged in as {resp_json['email']} using the browser."
            )
            _check_version_headers(resp)
            return
        # Fail for all other statuses
        consolidate_error(res=resp, description="Browser Login")
        # spinner.stop()

        return
    raise qnx_exc.AuthenticationError("Browser login Failed, code has expired.")


def login_with_credentials() -> None:
    """Log in to Nexus using a username and password."""
    user_name = input("Enter your Nexus email: ")
    pwd = getpass.getpass(prompt="Enter your Nexus password: ")

    _request_tokens(user=user_name, pwd=pwd)

    print(f"✅ Successfully logged in as {user_name}.")


def login_no_interaction(user: EmailStr, pwd: str) -> None:
    """Log in to Nexus using a username and password.
    Please be careful with storing credentials in plain text or source code.
    """
    _request_tokens(user=user, pwd=pwd)

    print(f"✅ Successfully logged in as {user}.")


def logout() -> None:
    """Clear tokens from file system and the client."""
    remove_token("refresh_token")
    remove_token("access_token")
    get_nexus_client(reload=True)
    print("Successfully logged out.")


def _request_tokens(user: EmailStr, pwd: str) -> None:
    """Method to send login request to Nexus auth api and save tokens."""
    body = {"email": user, "password": pwd}
    try:
        resp = _get_auth_client().post(
            "/login",
            json=body,
            headers={VERSION_HEADER: VERSION},
        )

        mfa_redirect_uri = resp.json().get("redirect_uri", "")
        if mfa_redirect_uri.startswith("/auth/mfa_challenge/"):
            mfa_code = input("Enter your MFA verification code: ")
            body["code"] = mfa_code
            body.pop("password")
            resp = _get_auth_client().post(
                "/mfa_challenge",
                json=body,
            )

        terms_redirect_uri = resp.json().get("redirect_uri", "")
        if terms_redirect_uri.startswith("/auth/terms_challenge"):
            message = "Terms and conditions not accepted. To continue, "
            message += "please accept our new terms and conditions by signing in "
            message += "to the Nexus website https://nexus.quantinuum.com/auth/login."

            # logger.error(message)
            raise qnx_exc.AuthenticationError(message)

        _response_check(resp, "Login")

        myqos_oat = resp.cookies.get("myqos_oat", None)
        myqos_id = resp.cookies.get("myqos_id", None)

        if not myqos_oat or not myqos_id:
            raise qnx_exc.AuthenticationError(
                "Authorization cookies missing from response."
            )

        write_token("refresh_token", myqos_oat)
        write_token("access_token", myqos_id)
        get_nexus_client(reload=True)

        _check_version_headers(resp)

    finally:
        del user
        del pwd
        del body


def _response_check(res: httpx.Response, description: str) -> None:
    """Consolidate as much error-checking of response"""
    # check if token has expired or is generally unauthorized
    resp_json = res.json()
    if res.status_code == HTTPStatus.UNAUTHORIZED:
        raise qnx_exc.AuthenticationError(
            (
                f"Authorization failure attempting: {description}."
                f"\n\nServer Response: {resp_json}"
            )
        )
    if res.status_code != HTTPStatus.OK:
        raise qnx_exc.AuthenticationError(
            f"HTTP error attempting: {description}.\n\nServer Response: {resp_json}"
        )



"""Tests related to running jobs against QSys devices."""

from typing import Any, Callable, ContextManager, cast

import pytest
from guppylang import guppy
from guppylang.std.builtins import result
from guppylang.std.quantum import cx, h, measure, qubit, x, z
from hugr.qsystem.result import QsysResult
from pytket.backends.backendinfo import BackendInfo
from quantinuum_schemas.models.backend_config import BasicEmulatorConfig

import qnexus as qnx
from qnexus.models.references import HUGRRef, ProjectRef, ResultVersions


def prepare_teleportation() -> Any:
    """Prepares the teleportation circuit."""

    @guppy
    def bell() -> tuple[qubit, qubit]:
        """Constructs a bell state."""
        q1, q2 = qubit(), qubit()
        h(q1)
        cx(q1, q2)
        return q1, q2

    @guppy
    def main() -> None:
        src = qubit()
        x(src)
        alice, bob = bell()

        cx(src, alice)
        h(src)
        if measure(alice):
            x(bob)
        if measure(src):
            z(bob)

        result("teleported", measure(bob))

    main.check()
    return main.compile()


@pytest.mark.skip(reason="Skipped while Selene is disabled for refactoring work.")
def test_guppy_execution(
    test_case_name: str,
    create_project: Callable[[str], ContextManager[ProjectRef]],
) -> None:
    """Test the execution of a guppy program
    on a next-generation QSys device."""

    with create_project(f"project for {test_case_name}") as project_ref:
        n_qubits = 3
        n_shots = 10

        hugr_ref = qnx.hugr.upload(
            hugr_package=prepare_teleportation(),
            name=f"hugr for {test_case_name}",
            project=project_ref,
        )

        job_ref = qnx.start_execute_job(
            programs=[hugr_ref],
            n_shots=[n_shots],
            backend_config=BasicEmulatorConfig(n_qubits=n_qubits),
            project=project_ref,
            name=f"basicemulator selene job for {test_case_name}",
        )

        qnx.jobs.wait_for(job_ref)

        results = qnx.jobs.results(job_ref)

        assert len(results) == 1
        result_ref = results[0]

        assert isinstance(result_ref.download_backend_info(), BackendInfo)
        assert isinstance(result_ref.get_input(), HUGRRef)

        assert result_ref.get_input().id == hugr_ref.id

        qsys_result = cast(QsysResult, result_ref.download_result())
        assert len(qsys_result.results) == n_shots
        assert qsys_result.results[0].entries[0][0] == "teleported"
        assert qsys_result.results[0].entries[0][1] == 1

        # check some QsysResults functionality
        assert len(qsys_result.collated_counts().items()) > 0

        # Assert we can get the same result for v4 results
        v4_qsys_result = cast(
            QsysResult, result_ref.download_result(version=ResultVersions.RAW)
        )
        assert len(v4_qsys_result.results) == n_shots
        assert v4_qsys_result.results[0].entries[0][0] == "USER:BOOL:teleported"
        assert v4_qsys_result.results[0].entries[0][1] == 1

        # This doesn't seem to work with v4 results, thinks everything is going to be a bit.
        # assert len(qsys_result.collated_counts().items()) > 0



"""Test basic functionality relating to the gpu_decoder_config."""

from typing import Callable, ContextManager

import pandas as pd

import qnexus as qnx
from qnexus.models.job_status import JobStatusEnum
from qnexus.models.references import GpuDecoderConfigRef, Ref


def test_gpu_decoder_config_download(
    test_case_name: str,
    create_gpu_decoder_config_in_project: Callable[
        [str, str, str], ContextManager[GpuDecoderConfigRef]
    ],
    qa_gpu_decoder_config: str,
    test_ref_serialisation: Callable[[str, Ref], None],
) -> None:
    """Test that valid GPU decoder config can be extracted from an uploaded GPU decoder config."""

    project_name = f"project for {test_case_name}"
    gpu_decoder_config_name = f"gpu decoder config for {test_case_name}"

    with create_gpu_decoder_config_in_project(
        project_name,
        gpu_decoder_config_name,
        qa_gpu_decoder_config,
    ) as gpu_decoder_config_ref:
        assert isinstance(gpu_decoder_config_ref, GpuDecoderConfigRef)
        downloaded_gpu_decoder_config = (
            gpu_decoder_config_ref.download_gpu_decoder_config_contents()
        )

        assert downloaded_gpu_decoder_config == qa_gpu_decoder_config

        gpu_decoder_config_ref_by_id = qnx.gpu_decoder_configs.get(
            id=gpu_decoder_config_ref.id
        )
        test_ref_serialisation("gpu_decoder_config", gpu_decoder_config_ref_by_id)


def test_gpu_decoder_config_flow(
    test_case_name: str,
    create_gpu_decoder_config_in_project: Callable[
        [str, str, str], ContextManager[GpuDecoderConfigRef]
    ],
    qa_gpu_decoder_config: str,
    qa_qir_bitcode: bytes,
) -> None:
    """Test the flow for executing a program on an OG device with a GPU decoder config."""

    project_name = f"project for {test_case_name}"
    gpu_decoder_config_name = f"gpu decoder config for {test_case_name}"
    qir_name = f"qir for {test_case_name}"

    with (
        create_gpu_decoder_config_in_project(
            project_name,
            gpu_decoder_config_name,
            qa_gpu_decoder_config,
        ) as gpu_decoder_config_ref,
    ):
        proj_ref = qnx.projects.get(name_like=project_name)

        my_gpu_decoder_config_db_matches = qnx.gpu_decoder_configs.get_all()
        assert isinstance(my_gpu_decoder_config_db_matches.summarize(), pd.DataFrame)
        assert isinstance(next(my_gpu_decoder_config_db_matches), GpuDecoderConfigRef)

        gpu_decoder_config_ref = qnx.gpu_decoder_configs.get(
            name_like=gpu_decoder_config_name
        )
        assert isinstance(gpu_decoder_config_ref, GpuDecoderConfigRef)

        gpu_decoder_config_ref_2 = qnx.gpu_decoder_configs.get(
            id=gpu_decoder_config_ref.id
        )
        assert gpu_decoder_config_ref == gpu_decoder_config_ref_2

        qir_program_ref = qnx.qir.upload(
            qir=qa_qir_bitcode, name=qir_name, project=proj_ref
        )

        execute_job_ref = qnx.start_execute_job(
            programs=[qir_program_ref],
            name=f"qir gpu decoder config execute job for {test_case_name}",
            n_shots=[100],
            backend_config=qnx.QuantinuumConfig(
                device_name="H1-1E",
                max_cost=10,
            ),
            gpu_decoder_config=gpu_decoder_config_ref,
            project=proj_ref,
        )

        # Don't fully submit the job (July 2025)
        qnx.jobs.wait_for(execute_job_ref, wait_for_status=JobStatusEnum.SUBMITTED)
        qnx.jobs.cancel(execute_job_ref)
        qnx.jobs.wait_for(execute_job_ref, wait_for_status=JobStatusEnum.CANCELLED)



"""Test basic functionality relating to the credential module."""

import qnexus as qnx
from qnexus.models import Credential, CredentialIssuer


def test_credential_get_all(
    authenticated_nexus: None,
) -> None:
    """Test that we can get a list of credentials."""
    creds = qnx.credentials.get_all()
    assert isinstance(creds, list)
    for cred in creds:
        assert isinstance(cred, Credential)
        assert isinstance(cred.backend_issuer, CredentialIssuer)



"""Client API for gpu decoder configs in Nexus."""

import base64
from datetime import datetime
from typing import Any, Union, cast
from uuid import UUID

import qnexus.exceptions as qnx_exc
from qnexus.client import get_nexus_client
from qnexus.client.nexus_iterator import NexusIterator
from qnexus.client.utils import handle_fetch_errors
from qnexus.context import (
    get_active_project,
    merge_project_from_context,
    merge_properties_from_context,
)
from qnexus.models.annotations import Annotations, CreateAnnotations, PropertiesDict
from qnexus.models.filters import (
    CreatorFilter,
    FuzzyNameFilter,
    PaginationFilter,
    ProjectRefFilter,
    PropertiesFilter,
    ScopeFilter,
    ScopeFilterEnum,
    SortFilter,
    SortFilterEnum,
    TimeFilter,
)
from qnexus.models.references import DataframableList, GpuDecoderConfigRef, ProjectRef


class Params(
    ScopeFilter,
    SortFilter,
    PaginationFilter,
    FuzzyNameFilter,
    CreatorFilter,
    ProjectRefFilter,
    PropertiesFilter,
    TimeFilter,
):
    """Params for filtering gpu decoder configs."""


@merge_project_from_context
def get_all(
    name_like: str | None = None,
    creator_email: list[str] | None = None,
    project: ProjectRef | None = None,
    properties: PropertiesDict | None = None,
    created_before: datetime | None = None,
    created_after: datetime | None = datetime(day=1, month=1, year=2023),
    modified_before: datetime | None = None,
    modified_after: datetime | None = None,
    sort_filters: list[SortFilterEnum] | None = None,
    page_number: int | None = None,
    page_size: int | None = None,
    scope: ScopeFilterEnum | None = None,
) -> NexusIterator[GpuDecoderConfigRef]:
    """Get a NexusIterator over gpu decoder configs with optional filters."""

    params = Params(
        name_like=name_like,
        creator_email=creator_email,
        properties=properties,
        project=project,
        created_before=created_before,
        created_after=created_after,
        modified_before=modified_before,
        modified_after=modified_after,
        sort=SortFilter.convert_sort_filters(sort_filters),
        page_number=page_number,
        page_size=page_size,
        scope=scope,
    ).model_dump(by_alias=True, exclude_unset=True, exclude_none=True)

    return NexusIterator(
        resource_type="GpuDecoderConfig",
        nexus_url="/api/gpu_decoder_configs/v1beta",
        params=params,
        wrapper_method=_to_gpu_decoder_config_ref,
        nexus_client=get_nexus_client(),
    )


def _to_gpu_decoder_config_ref(
    page_json: dict[str, Any],
) -> DataframableList[GpuDecoderConfigRef]:
    """Convert JSON response dict to a list of GpuDecoderConfigRefs."""

    gpu_decoder_config_refs: DataframableList[GpuDecoderConfigRef] = DataframableList(
        []
    )

    for gpu_decoder_config_data in page_json["data"]:
        project_id = gpu_decoder_config_data["relationships"]["project"]["data"]["id"]
        project_details = next(
            proj for proj in page_json["included"] if proj["id"] == project_id
        )
        project = ProjectRef(
            id=project_id,
            annotations=Annotations.from_dict(project_details["attributes"]),
            contents_modified=project_details["attributes"]["contents_modified"],
            archived=project_details["attributes"]["archived"],
        )

        gpu_decoder_config_refs.append(
            GpuDecoderConfigRef(
                id=UUID(gpu_decoder_config_data["id"]),
                annotations=Annotations.from_dict(
                    gpu_decoder_config_data["attributes"]
                ),
                project=project,
            )
        )
    return gpu_decoder_config_refs


def get(
    *,
    id: Union[UUID, str, None] = None,
    name_like: str | None = None,
    creator_email: list[str] | None = None,
    project: ProjectRef | None = None,
    properties: PropertiesDict | None = None,
    created_before: datetime | None = None,
    created_after: datetime | None = datetime(day=1, month=1, year=2023),
    modified_before: datetime | None = None,
    modified_after: datetime | None = None,
    sort_filters: list[SortFilterEnum] | None = None,
    page_number: int | None = None,
    page_size: int | None = None,
    scope: ScopeFilterEnum | None = None,
) -> GpuDecoderConfigRef:
    """
    Get a single gpu_decoder_config using filters. Throws an exception if the filters do
    not match exactly one object.
    """
    if id:
        return _fetch_by_id(gpu_decoder_config_id=id, scope=scope)

    return get_all(
        name_like=name_like,
        creator_email=creator_email,
        properties=properties,
        project=project,
        created_before=created_before,
        created_after=created_after,
        modified_before=modified_before,
        modified_after=modified_after,
        sort_filters=sort_filters,
        page_number=page_number,
        page_size=page_size,
        scope=scope,
    ).try_unique_match()


@merge_properties_from_context
def upload(
    gpu_decoder_config: str,
    project: ProjectRef | None = None,
    name: str | None = None,
    description: str | None = None,
    properties: PropertiesDict | None = None,
) -> GpuDecoderConfigRef:
    """Upload a pytket GpuDecoderConfig to Nexus."""
    project = project or get_active_project(project_required=True)
    project = cast(ProjectRef, project)

    attributes = {"contents": base64.b64encode(gpu_decoder_config.encode()).decode()}
    if name is None:
        raise ValueError("GpuDecoderConfig must have a name to be uploaded")

    annotations = CreateAnnotations(
        name=name,
        description=description,
        properties=properties,
    ).model_dump(exclude_none=True)
    attributes.update(annotations)
    relationships = {"project": {"data": {"id": str(project.id), "type": "project"}}}

    req_dict = {
        "data": {
            "attributes": attributes,
            "relationships": relationships,
            "type": "gpu_decoder_config",
        }
    }

    res = get_nexus_client().post("/api/gpu_decoder_configs/v1beta", json=req_dict)

    # https://cqc.atlassian.net/browse/MUS-3054
    if res.status_code != 201:
        raise qnx_exc.ResourceCreateFailed(
            message=res.text, status_code=res.status_code
        )

    res_data_dict = res.json()["data"]

    return GpuDecoderConfigRef(
        id=UUID(res_data_dict["id"]),
        annotations=Annotations.from_dict(res_data_dict["attributes"]),
        project=project,
    )


@merge_properties_from_context
def update(
    ref: GpuDecoderConfigRef,
    name: str | None = None,
    description: str | None = None,
    properties: PropertiesDict | None = None,
) -> GpuDecoderConfigRef:
    """Update the annotations on a GpuDecoderConfigRef."""
    ref_annotations = ref.annotations.model_dump()
    annotations = Annotations(
        name=name,
        description=description,
        properties=properties if properties else PropertiesDict(),
    ).model_dump(exclude_none=True)
    ref_annotations.update(annotations)

    req_dict = {
        "data": {
            "attributes": annotations,
            "relationships": {},
            "type": "gpu_decoder_config",
        }
    }

    res = get_nexus_client().patch(
        f"/api/gpu_decoder_configs/v1beta/{ref.id}", json=req_dict
    )

    if res.status_code != 200:
        raise qnx_exc.ResourceUpdateFailed(
            message=res.text, status_code=res.status_code
        )

    res_dict = res.json()["data"]

    return GpuDecoderConfigRef(
        id=UUID(res_dict["id"]),
        annotations=Annotations.from_dict(res_dict["attributes"]),
        project=ref.project,
    )


def _fetch_by_id(
    gpu_decoder_config_id: UUID | str, scope: ScopeFilterEnum | None
) -> GpuDecoderConfigRef:
    """Utility method for fetching directly by a unique identifier."""
    params = Params(
        scope=scope,
    ).model_dump(by_alias=True, exclude_unset=True, exclude_none=True)

    res = get_nexus_client().get(
        f"/api/gpu_decoder_configs/v1beta/{gpu_decoder_config_id}", params=params
    )

    handle_fetch_errors(res)

    res_dict = res.json()

    project_id = res_dict["data"]["relationships"]["project"]["data"]["id"]
    project_details = next(
        proj for proj in res_dict["included"] if proj["id"] == project_id
    )
    project = ProjectRef(
        id=project_id,
        annotations=Annotations.from_dict(project_details["attributes"]),
        contents_modified=project_details["attributes"]["contents_modified"],
        archived=project_details["attributes"]["archived"],
    )

    return GpuDecoderConfigRef(
        id=UUID(res_dict["data"]["id"]),
        annotations=Annotations.from_dict(res_dict["data"]["attributes"]),
        project=project,
    )


def _fetch_gpu_decoder_config(handle: GpuDecoderConfigRef) -> str:
    """Utility method for fetching a gpu decoder config from a GpuDecoderConfigRef."""
    res = get_nexus_client().get(f"/api/gpu_decoder_configs/v1beta/{handle.id}")
    if res.status_code != 200:
        raise qnx_exc.ResourceFetchFailed(message=res.text, status_code=res.status_code)

    return base64.b64decode(res.json()["data"]["attributes"]["contents"]).decode()



"""Test basic functionality relating to setting 'scope' for API requests."""

import qnexus as qnx
from qnexus.models.filters import ScopeFilterEnum


def test_user_scope(
    authenticated_nexus: None,
) -> None:
    """Test that we can add a scope filter to our queries,
    but that this is ignored when the user claims to be an
    org_admin or global_admin."""

    all_my_user_projects = qnx.projects.get_all(scope=ScopeFilterEnum.USER)

    assert all_my_user_projects.count() > 1

    all_org_admin_projects = qnx.projects.get_all(scope=ScopeFilterEnum.ORG_ADMIN)

    assert all_org_admin_projects.count() == 0

    all_global_admin_projects = qnx.projects.get_all(scope=ScopeFilterEnum.GLOBAL_ADMIN)

    assert all_global_admin_projects.count() == 0

    all_highest_scope_projects = qnx.projects.get_all(scope=ScopeFilterEnum.HIGHEST)

    assert all_highest_scope_projects.count() == all_my_user_projects.count()



"""Test that we can set properties on a project, and use them when creating/filtering resources."""

from typing import Callable, ContextManager, OrderedDict

from pytket.circuit import Circuit

import qnexus as qnx
from qnexus.models.annotations import PropertiesDict
from qnexus.models.references import ProjectRef


def test_property_creation_and_filtering(
    test_case_name: str,
    create_project: Callable[[str], ContextManager[ProjectRef]],
) -> None:
    """Test that we can create resources with property values
    and filter resources by their property values."""

    project_name = f"project for {test_case_name}"

    with create_project(project_name) as my_new_project:
        test_property_name_1 = "QA_test_prop_1"
        test_property_name_2 = "QA_test_prop_2"
        test_property_name_3 = "QA_test_prop_3"
        test_property_name_4 = "QA_test_prop_4"

        qnx.projects.add_property(
            name=test_property_name_1,
            property_type="bool",
            project=my_new_project,
            description="A test property for my QA test project",
        )

        qnx.projects.add_property(
            name=test_property_name_2,
            property_type="string",
            project=my_new_project,
            description="Another test property for my QA test project",
        )

        qnx.projects.add_property(
            name=test_property_name_3,
            property_type="int",
            project=my_new_project,
            description="Another test property for my QA test project",
        )

        qnx.projects.add_property(
            name=test_property_name_4,
            property_type="float",
            project=my_new_project,
            description="Another test property for my QA test project",
        )

        test_props = qnx.projects.get_properties(my_new_project)

        assert len(test_props) == 4

        circuit_name = f"circuit for {test_case_name}"

        my_circ = Circuit(2, 2).H(0).CX(0, 1).measure_all()

        qnx.circuits.upload(
            circuit=my_circ,
            name=circuit_name,
            project=my_new_project,
            properties=OrderedDict(
                {
                    test_property_name_1: True,
                    test_property_name_2: "test_string",
                    test_property_name_3: 42,
                    test_property_name_4: 3.15,
                }
            ),
        )

        qnx.circuits.upload(
            circuit=my_circ,
            name=circuit_name,
            project=my_new_project,
            properties=OrderedDict(
                {
                    test_property_name_1: False,
                }
            ),
        )

        my_circuit_refs = qnx.circuits.get_all(
            project=my_new_project,
            properties=PropertiesDict(
                {
                    test_property_name_1: True,
                    test_property_name_2: "test_string",
                    test_property_name_3: 42,
                    test_property_name_4: 3.15,
                }
            ),
        )

        assert my_circuit_refs.count() == 1
        my_circuit_ref = my_circuit_refs.list()[0]

        assert my_circuit_ref.annotations.properties[test_property_name_1] is True
        assert (
            my_circuit_ref.annotations.properties[test_property_name_2] == "test_string"
        )



"""Test basic functionality relating to the hugr module."""

from typing import Callable, ContextManager

import pytest
from hugr.package import Package

import qnexus as qnx
import qnexus.exceptions as qnx_exc
from qnexus.models.annotations import PropertiesDict
from qnexus.models.references import HUGRRef, ProjectRef, Ref


def test_hugr_create_and_update(
    test_case_name: str,
    qa_hugr_package: Package,
    create_property_in_project: Callable[..., ContextManager[ProjectRef]],
) -> None:
    """Test that we can create a hugr and add a property value."""
    project_name = f"project for {test_case_name}"
    property_name = f"property for {test_case_name}"

    with create_property_in_project(
        project_name=project_name,
        property_name=property_name,
        property_type="string",
        required=False,
    ) as my_proj:
        hugr_name = f"hugr for {test_case_name}"
        my_new_hugr = qnx.hugr.upload(
            hugr_package=qa_hugr_package, name=hugr_name, project=my_proj
        )

        assert isinstance(my_new_hugr, HUGRRef)

        test_prop_value = "foo"
        updated_hugr_ref = qnx.hugr.update(
            ref=my_new_hugr,
            properties=PropertiesDict({property_name: test_prop_value}),
        )

        assert updated_hugr_ref.annotations.properties[property_name] == test_prop_value


def test_hugr_download(
    test_case_name: str,
    qa_hugr_package: Package,
    create_hugr_in_project: Callable[[str, str, Package], ContextManager[HUGRRef]],
) -> None:
    """Test that valid HUGR can be extracted from an uploaded HUGR module."""
    project_name = f"project for {test_case_name}"
    hugr_name = f"hugr for {test_case_name}"

    with create_hugr_in_project(
        project_name,
        hugr_name,
        qa_hugr_package,
    ) as hugr_ref:
        with pytest.raises(qnx_exc.ResourceFetchFailed):
            # Temporarily disabled due to missing functionality in hugr
            downloaded_hugr_package = hugr_ref.download_hugr()
            assert isinstance(downloaded_hugr_package, Package)
            assert qa_hugr_package == downloaded_hugr_package
        downloaded_hugr_bytes = hugr_ref.download_hugr_bytes()
        assert isinstance(downloaded_hugr_bytes, bytes)
        assert qa_hugr_package.to_bytes() == downloaded_hugr_bytes


def test_hugr_get_by_id(
    test_case_name: str,
    qa_hugr_package: Package,
    create_hugr_in_project: Callable[[str, str, Package], ContextManager[HUGRRef]],
    test_ref_serialisation: Callable[[str, Ref], None],
) -> None:
    """Test that we can get a HUGRRef by its ID and the HUGRRef
    serialisation round trip."""
    project_name = f"project for {test_case_name}"
    hugr_name = f"hugr for {test_case_name}"

    with create_hugr_in_project(
        project_name,
        hugr_name,
        qa_hugr_package,
    ):
        my_proj = qnx.projects.get(name_like=project_name)
        my_hugr_ref = qnx.hugr.get(name_like=hugr_name, project=my_proj)

        hugr_ref_by_id = qnx.hugr.get(id=my_hugr_ref.id)

        assert hugr_ref_by_id == my_hugr_ref

        test_ref_serialisation("hugr", hugr_ref_by_id)


def test_hugr_get_all(
    test_case_name: str,
    qa_hugr_package: Package,
    create_hugr_in_project: Callable[[str, str, Package], ContextManager[HUGRRef]],
) -> None:
    """Test that we can get all HUGRRefs in a project."""
    project_name = f"project for {test_case_name}"
    hugr_name = f"hugr for {test_case_name}"

    with create_hugr_in_project(
        project_name,
        hugr_name,
        qa_hugr_package,
    ):
        my_proj = qnx.projects.get(name_like=project_name)

        hugrs = qnx.hugr.get_all(project=my_proj)

        assert hugrs.count() >= 1
        assert isinstance(hugrs.list()[0], HUGRRef)
